'''
## This file is autogenerated from macrosynergy setup.py.

It contains version information for the macrosynergy package.

short_version: 0.1.34

version: 0.1.34

full_version: 0.1.34

git_revision: 9746fc05091b47181286ef70ce480a5ece70d7fb

release: True

'''
short_version: str = '0.1.34'
version: str = '0.1.34'
full_version: str = '0.1.34'
git_revision: str = '9746fc05091b47181286ef70ce480a5ece70d7fb'
release: bool = True
if not release:
    version = full_version
